#include "utilities.h"
#include "appli1.h"

void CheckArgs(int argc, char* argv[])
{
    if (argc < 2) {
	fprintf(stderr, "Not enough parameters : %s <input file prefix> <output file name>\n", argv[0]);
	exit(1);
    }
}

void GoPmacro(MpegGopHdr* gopHdr, BitParser* obp, int *temporalRef)
{
	/* Setting GoP header to the right values*/
        mMpegGopHdrSet(gopHdr);
	/* And send it to the OutputBitParser */
        MpegGopHdrEncode(gopHdr, obp);
	/* temporalRef set to 0 at the beginning of each GoP */
        *temporalRef = 0;
}

void PicImacro(YUV_image* yuv, YUV_image* prev, MpegPicHdr *picHdr, Sc_image* scpr, int* sliceInfo, int sliceInfoLen, BitParser* obp, int* temporalRef)
{
	/* Setting PicHeader to the right values, saying 'I Frame' and giving the temporalRef */
	mMpegPicHdrSet(picHdr, 'I', *temporalRef);
	/* and send it to the OutputBitParser */
        MpegPicHdrEncode (picHdr, obp);

	/* Transform our YUV image in a DCT_space image scpr */
	mByteToScI(yuv, NULL, scpr);
	/* and encode it in the OutputBitParser according to the settings in picHdr, this header have already been encoded */
	mMpegPicIEncode(picHdr, scpr, sliceInfo, sliceInfoLen,obp);

	/* increment both time and temporalRef */
        IncrementTime();
        *temporalRef = *temporalRef +1;

	/* Now that we are done, just swap the current and the reference image */
	/* this way, next processing will write over the previous one and use the 
	   current one as previous */
        Swap(yuv, prev);
}

void PicPmacro(YUV_image* yuv, YUV_image* prev, MpegPicHdr *picHdr, Sc_image* scpr, int* sliceInfo, int sliceInfoLen, BitParser* obp, int* temporalRef)
{
	/* Setting PicHeader to the right values, saying 'P Frame' and giving the temporalRef */
	mMpegPicHdrSet(picHdr, 'P', *temporalRef);
	/* and send it to the OutputBitParser */
        MpegPicHdrEncode (picHdr, obp);

	/* Read yuv, and perform Motion-Estimation relative to prev. 
	   that is, for each macroblock in yuv, seek the most similar
	   one in prev and encode in scpr->fmv the corresponding vector */
	/* Create a difference Image locally : for each macroblock in 
	   yuv, encode it minus the corresponding translated macroblock
	   of the previous image */
	/* Transform the difference  image in a DCT_space image scpr */
	mByteToScP(picHdr, yuv, prev, scpr);
	/* encode the difference and the vectors and send it to the OutputBitParser */
	mMpegPicPEncode(picHdr,scpr,sliceInfo,sliceInfoLen,obp);

        /* increment both time and temporalRef */
        IncrementTime();
        *temporalRef = *temporalRef +1;

	/* Swap - same thing as in the I-image */
        Swap(yuv, prev);
}

void WritingFiles(BitStream* obs, BitParser* obp, FILE* outFile)
{
	/* Flush all the contents of the OutputBitParser and
	   write them in outFile */
        BitStreamFileWriteSegment(obs, outFile, 0, BitParserTell(obp));
	/* re-attach outputBitParser and outputBitStream
	   before we use these again */
        BitParserWrap(obp, obs);
}

void READ_TO_YUV(BitParser *bp, BitStream *bs, RGB_image* rgb, YUV_image* yuv, char* namePrefix, PnmHdr *pnmHdr) 
{
    FILE* inFile;
    char fileName[PREFIX_SIZE+3];

    sprintf(fileName, "%s%03d.ppm", namePrefix, pictures());
    printf("Processing %s\n", fileName);
    inFile = fopen(fileName, "rb");
    if (inFile == NULL) {
        fprintf(stderr, "unable to open %s for reading.\n", fileName);
        exit(1);
    }
    BitParserWrap(bp, bs);
    BitStreamFileRead(bs, inFile, 0);
    PnmHdrParse(bp, pnmHdr);
    PpmParse(bp, rgb->r, rgb->g, rgb->b);
    fclose(inFile);
    RgbToYuv420(rgb->r, rgb->g, rgb->b, yuv->y, yuv->u, yuv->v);
}


int main(int argc, char *argv[])
{
    int w = WIDTH, h = HEIGHT;
    int sliceInfo[] = {1000};       /* this should be enough for one frame */
    int sliceInfoLen = 1;
    int gop, pic, temporalRef;
    FILE *inFile, *outFile;
    char fileName[PREFIX_SIZE+3];
    char namePrefix[PREFIX_SIZE];

    /* input and output mechanisms */
    /* A BitStream is a memory buffer */
    /* A BitParser is a set of tools to read/write using the BitStream */
    BitParser *bp  = BitParserNew();        /* input bitstream */
    BitStream *bs;
    BitParser *obp = BitParserNew();        /* output bitstream */
    BitStream *obs = BitStreamNew(BUFFER_SIZE);

    /* Temporary datas : a RGB image used when parsing PPM files */
    /* yuv is the YUV conversion of rgb, prev is the last yuv used before the current one */
    /* scpr is a SemiCompressed image, ie an image encoded either
	- as a JPEG image (DCT domain), ie for I-frames
	- as a P frame, that the DCT image corresponding to the image minus a prediction */
    /* prediction is described as the set of vectors fmv. */
    /* For each macroblock, the prediction is prev translated by the corresponding vector */
    RGB_image* rgb = RGB_init(w, h);
    YUV_image* yuv = YUV_init(w, h);
    YUV_image* prev = YUV_init(w, h);
    Sc_image* scpr = Sc_init(w, h);

    /* Allocate memory for all the headers */
    /* By the way, we already set the sequence Header */
    PnmHdr *pnmHdr = PnmHdrNew();
    MpegSeqHdr *seqHdr = MpegSeqHdrInit(w, h);
    MpegGopHdr *gopHdr = MpegGopHdrNew();
    MpegPicHdr *picHdr = MpegPicHdrNew();


    /* Check arguments... */
    CheckArgs(argc, argv);

    /* Get input file size */
    /********************************************************************************/
    /***                                                                          ***/
    /***                        Add your own code here !                          ***/
    /***                                                                          ***/
    /*** Check if the first .ppm file exists and be sure its name is in fileName  ***/
    /***                                                                          ***/
    /********************************************************************************/



    /* Create bs as a BitStream (read buffer) of the size of an input file */
    bs  = BitStreamNew(fsize(fileName));

    /* Create outfile file */
    /********************************************************************************/
    /***                                                                          ***/
    /***                        Add your own code here !                          ***/
    /***                                                                          ***/
    /***     Open and check outFile - its path is given by the second argument    ***/
    /***                                                                          ***/
    /********************************************************************************/



    /* Attach the output BitParser to the output BitStream */
    BitParserWrap(obp, obs);


    /* Send the Sequence Header to the OutputBitParser */
    /* Values have already been set by MpegSeqHdrInit */
    MpegSeqHdrEncode(seqHdr, obp);

    for (gop = 0; gop < MOV_LEN/GOP_LEN; gop++) {
	/* Init the GoPHeader and send it to the OutputBitParser */
	GoPmacro(gopHdr, obp, &temporalRef);

        /* I FRAME */
	/* Read the current frame and write it in YUV format in the YUV structure */
        READ_TO_YUV(bp, bs, rgb, yuv, namePrefix, pnmHdr);
	/* Encoding the image in yuv as an Intra Image and send it to the OutputBitParser */
	/* swap the current image yuv and the previous one prev */
	PicImacro(yuv, prev, picHdr, scpr, sliceInfo, sliceInfoLen, obp, &temporalRef);

            /* P FRAME */
        for (pic = 1; pic < GOP_LEN ; pic++) {
	    /* read current image and store in yuv */
            READ_TO_YUV(bp, bs, rgb, yuv, namePrefix, pnmHdr);
	    /* Encoding yuv as a fonction of prev, send to OutputBitParser and swap yuv and prev */
	    PicPmacro(yuv, prev, picHdr, scpr, sliceInfo, sliceInfoLen, obp, &temporalRef);
	    /* Write all the content of OutputBitParser in outFile */
	    WritingFiles(obs, obp, outFile);
        }
    }
    /* Sending the official Mpeg Sequence End Code to obp */
    MpegSeqEndCodeEncode(obp);
    /* and conclude the writing */
    WritingFiles(obs, obp, outFile);
 
 
    /* lots of things to free up */
    BitStreamFree(bs);
    BitParserFree(bp);
    BitStreamFree(obs);
    BitParserFree(obp);

    MpegSeqHdrFree(seqHdr);
    MpegGopHdrFree(gopHdr);
    MpegPicHdrFree(picHdr);
    PnmHdrFree(pnmHdr);

    RGB_free(rgb);
    YUV_free(yuv);
    YUV_free(prev);
    Sc_free(scpr);

    return 0;
}
